# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.project.cplusplus** registers a ``"C++"`` projects support.
"""

from proj_flow.project.cplusplus import cmake_context, conan_context

__all__ = ["cmake_context", "conan_context"]
