# Copyright (c) 2025 Marcin Zdun
# This code is licensed under MIT license (see LICENSE for details)

"""
The **proj_flow.ext.github** provides GitHub support through CLI and GitHub
hosting plugin.
"""

from . import cli, hosting, switches

__all__ = ["cli", "hosting", "switches"]
